
/*
 * MQTT Code for nRF9160 Demo
 *
 * Copyright (c) M2MD Technologies 2019
 * All Rights Reserved
 */

#include <zephyr.h>
#include <kernel.h>
#include <logging/log.h>
#include <net/socket.h>
#include <net/mqtt.h>
#include <lte_lc.h>
#include <nrf_errno.h>
#include <nrf_inbuilt_key.h>

#include "certificates.h"
#include "common.h"

#define PSK_SEC_TAG 0x555
#define CERT_SEC_TAG 0x111

/* Buffers for MQTT client. */
static u8_t rx_buffer[1024]; 
static u8_t tx_buffer[1024]; 
static u8_t payload_buf[1024];

/* The mqtt client struct */
static struct mqtt_client client;

/* MQTT Broker details. */
static struct sockaddr_storage broker;

/* Connected flag */
static bool connected;

/* File descriptor */
static struct pollfd fds;

LOG_MODULE_REGISTER(demo_mqtt, LOG_LEVEL_DBG);

/**@brief Configures modem to provide LTE link. Blocks until link is
 * successfully established.
 */
int modem_configure(void)
{
        int err = 0;
#if defined(CONFIG_LTE_LINK_CONTROL)
    if (IS_ENABLED(CONFIG_LTE_AUTO_INIT_AND_CONNECT)) {
        /* Do nothing, modem is already turned on
         * and connected.
         */
    } else {

        //k_sleep(K_MSEC(5000));
        LOG_DBG("LTE Link Connecting ...");
        err = lte_lc_init_and_connect();
        if(err != 0) {
            LOG_ERR("LTE link could not be established");
        }
        else {
            LOG_INF("LTE Link Connected!");
        }
    }

#endif
    return err;
}

/**@brief Function to print strings without null-termination
 */
static void data_print(u8_t *prefix, u8_t *data, size_t len)
{
    char buf[len + 1];

    memcpy(buf, data, len);
    buf[len] = 0;
    printk("%s%s\n", prefix, buf);
    //LOG_INF("%s%s", prefix, buf); this fails becasue strings are in memory not accessible by the logger
}

/**@brief Function to publish data on the configured topic
 */
static int data_publish(struct mqtt_client *c, enum mqtt_qos qos,
                        u8_t *data, size_t len)
{
    struct mqtt_publish_param param;
#define PUB_TOPIC \
    CONFIG_DEMO_MQTT_FLEET"/" \
    CONFIG_DEMO_MQTT_CLIENT_ID"/" \
    CONFIG_DEMO_MQTT_PUB_TOPIC

    param.message.topic.qos = qos;
    param.message.topic.topic.utf8 = PUB_TOPIC;
    param.message.topic.topic.size = strlen(PUB_TOPIC);
    param.message.payload.data = data;
    param.message.payload.len = len;
    param.message_id = sys_rand32_get();
    param.dup_flag = 0;
    param.retain_flag = 0;

    data_print("Publishing: ", data, len);
    LOG_INF("to topic: %s len: %u",
            PUB_TOPIC,
            (unsigned int)strlen(PUB_TOPIC));

    return mqtt_publish(c, &param);
}

/**@brief Function to subscribe to the configured topic
 */
static int subscribe(void)
{
#define SUB_TOPIC \
    CONFIG_DEMO_MQTT_FLEET"/" \
    CONFIG_DEMO_MQTT_CLIENT_ID"/" \
    CONFIG_DEMO_MQTT_SUB_TOPIC

    struct mqtt_topic subscribe_topic = {
        .topic = {
            .utf8 = SUB_TOPIC,
            .size = strlen(SUB_TOPIC)
        },
        .qos = MQTT_QOS_1_AT_LEAST_ONCE
    };

    const struct mqtt_subscription_list subscription_list = {
        .list = &subscribe_topic,
        .list_count = 1,
        .message_id = 1234
    };

    LOG_INF("Subscribing to: %s len %u", SUB_TOPIC,
            (unsigned int)strlen(SUB_TOPIC));

    return mqtt_subscribe(&client, &subscription_list);
}

/**@brief Function to read the published payload.
 */
static int publish_get_payload(struct mqtt_client *c, size_t length)
{
    u8_t *buf = payload_buf;
    u8_t *end = buf + length;

    if (length > sizeof(payload_buf)) {
        return -EMSGSIZE;
    }

    while (buf < end) {
        int ret = mqtt_read_publish_payload(c, buf, end - buf);

        if (ret < 0) {
            int err;

            if (ret != -EAGAIN) {
                return ret;
            }

            LOG_INF("mqtt_read_publish_payload: EAGAIN");

            err = poll(&fds, 1, K_SECONDS(CONFIG_MQTT_KEEPALIVE));
            if (err > 0 && (fds.revents & POLLIN) == POLLIN) {
                continue;
            } else {
                return -EIO;
            }
        }

        if (ret == 0) {
            return -EIO;
        }

        buf += ret;
    }

    return 0;
}

/**@brief MQTT client event handler
 */
void mqtt_evt_handler(struct mqtt_client *const c,
                      const struct mqtt_evt *evt)
{
    int err;

    switch (evt->type) {
    case MQTT_EVT_CONNACK:
        if (evt->result != 0) {
            LOG_INF("MQTT connect failed %d", evt->result);
            break;
        }

        connected = true;
        LOG_INF("[%s:%d] MQTT client connected!", __func__, __LINE__);
        subscribe();
        break;

    case MQTT_EVT_DISCONNECT:
        LOG_INF("[%s:%d] MQTT client disconnected %d", __func__,
                __LINE__, evt->result);

        connected = false;
        break;

    case MQTT_EVT_PUBLISH: {
        const struct mqtt_publish_param *p = &evt->param.publish;

        LOG_INF("[%s:%d] MQTT PUBLISH result=%d len=%d", __func__,
                __LINE__, evt->result, p->message.payload.len);
        err = publish_get_payload(c, p->message.payload.len);
        if (err >= 0) {
            data_print("Received: ", payload_buf,
                       p->message.payload.len);

            int len = p->message.payload.len;
            char buf[len + 1];
            memcpy(buf, payload_buf, len);
            buf[len] = 0;

            /* Match message with application-specific commands
             * and handle them

             ...
             */

        } else {
            LOG_INF("mqtt_read_publish_payload: Failed! %d", err);
            LOG_INF("Disconnecting MQTT client...");

            err = mqtt_disconnect(c);
            if (err) {
                LOG_INF("Could not disconnect: %d", err);
            }
        }
    }
    break;

    case MQTT_EVT_PUBACK:
        if (evt->result != 0) {
            LOG_INF("MQTT PUBACK error %d", evt->result);
            break;
        }

        LOG_INF("[%s:%d] PUBACK packet id: %u", __func__, __LINE__,
                evt->param.puback.message_id);
        break;

    case MQTT_EVT_SUBACK:
        if (evt->result != 0) {
            LOG_INF("MQTT SUBACK error %d", evt->result);
            break;
        }

        LOG_INF("[%s:%d] SUBACK packet id: %u", __func__, __LINE__,
                evt->param.suback.message_id);
        break;

    default:
        LOG_INF("[%s:%d] default: %d", __func__, __LINE__,
                evt->type);
        break;
    }
}

/**@brief Resolves the configured hostname and
 * initializes the MQTT broker structure
 */
static void broker_init(void)
{
    int err;

    struct addrinfo *result;
    struct addrinfo *addr;
    struct addrinfo hints = {
        .ai_family = AF_INET,
        .ai_socktype = SOCK_STREAM
    };

    LOG_INF("Getting addrinfo for -->%s<--", CONFIG_DEMO_MQTT_BROKER_HOSTNAME);
    err = getaddrinfo(CONFIG_DEMO_MQTT_BROKER_HOSTNAME, NULL, &hints, &result);
    if (err) {
        LOG_INF("ERROR: getaddrinfo failed %d", err);

        return;
    }

    addr = result;
    err = -ENOENT;
    /* Look for address of the broker. */
    while (addr != NULL) {
        /* IPv4 Address. */
        LOG_INF("LOOOOOPING\n");
        if (addr->ai_addrlen == sizeof(struct sockaddr_in)) {
            struct sockaddr_in *broker4 =
                ((struct sockaddr_in *)&broker);
            char ipv4_addr[NET_IPV4_ADDR_LEN];

            broker4->sin_addr.s_addr =
                ((struct sockaddr_in *)addr->ai_addr)
                ->sin_addr.s_addr;
            broker4->sin_family = AF_INET;

            uint16_t broker_port = CONFIG_DEMO_MQTT_BROKER_PORT;
            LOG_INF("Broker port is %u", broker_port);

            broker4->sin_port = htons(broker_port);

            inet_ntop(AF_INET, &broker4->sin_addr.s_addr,
                      ipv4_addr, sizeof(ipv4_addr));

            break;
        } else {
            LOG_INF("ai_addrlen = %u should be %u or %u",
                    (unsigned int)addr->ai_addrlen,
                    (unsigned int)sizeof(struct sockaddr_in),
                    (unsigned int)sizeof(struct sockaddr_in6));
        }

        addr = addr->ai_next;
        break;
    }

    /* Free the address. */
    freeaddrinfo(result);
}

/* Provisioned PSK here, redacted for dev zone  */
static uint8_t test_psk[64] = { 0 };

static int tls_enabled = 1;
static int psk_enabled = 1;

static const char *nrf_err_str(int err)
{
    switch(err) {
        case NRF_EFAULT:
            return "NRF_EFAULT";
        case NRF_EIO:
            return "NRF_EIO";
        case NRF_ENOBUFS:
            return "NRF_ENOBUFS";
        case NRF_ENOENT:
            return "NRF_ENOENT";
        case NRF_ENOMEM:
            return "NRF_ENOMEM";
        case NRF_EPERM:
            return "NRF_EPERM";
        case NRF_EACCES:
            return "NRF_EACCES";
        case NRF_EINVAL:
            return "NRF_EINVAL";
    }

    return "NOMATCH";
}

static int provision_certificates(nrf_sec_tag_t sec_tag)
{
    int err;

    for (nrf_key_mgnt_cred_type_t type = 0; type < 5; type++) {
        err = nrf_inbuilt_key_delete(sec_tag, type);
        printk("nrf_inbuilt_key_delete(%u, %d) => result=%d\n",
                sec_tag, type, err);
    }

    /* Provision CA Certificate. */
    err = nrf_inbuilt_key_write(sec_tag,
            NRF_KEY_MGMT_CRED_TYPE_CA_CHAIN,
            CLOUD_CA_CERTIFICATE,
            strlen(CLOUD_CA_CERTIFICATE));
    printk("nrf_inbuilt_key_write => result=%d\n", err);
    if (err) {
        printk("CLOUD_CA_CERTIFICATE err: %d", err);
        return err;
    }

    /* Provision Private Certificate. */
    err = nrf_inbuilt_key_write(
            sec_tag,
            NRF_KEY_MGMT_CRED_TYPE_PRIVATE_CERT,
            CLOUD_CLIENT_PRIVATE_KEY,
            strlen(CLOUD_CLIENT_PRIVATE_KEY));
    printk("nrf_inbuilt_key_write => result=%d\n", err);
    if (err) {
        printk("NRF_CLOUD_CLIENT_PRIVATE_KEY err: %d", err);
        return err;
    }

    /* Provision Public Certificate. */
    err = nrf_inbuilt_key_write(
            sec_tag,
            NRF_KEY_MGMT_CRED_TYPE_PUBLIC_CERT,
            CLOUD_CLIENT_PUBLIC_CERTIFICATE,
            strlen(CLOUD_CLIENT_PUBLIC_CERTIFICATE));
    printk("nrf_inbuilt_key_write => result=%d\n", err);
    if (err) {
        printk("CLOUD_CLIENT_PUBLIC_CERTIFICATE err: %d",
                err);
        return err;
    }

    return 0;
}

static int provision_psk_nrf(nrf_sec_tag_t sec_tag)
{
    int err = 0;

    for (nrf_key_mgnt_cred_type_t type = 0; type < 5; type++) {
        err = nrf_inbuilt_key_delete(sec_tag, type);
        LOG_INF("nrf_inbuilt_key_delete(%X, %d) => result=%d\n",
                sec_tag, type, err);
    }

    /* Provision Private Certificate. */
    err = nrf_inbuilt_key_write(sec_tag, NRF_KEY_MGMT_CRED_TYPE_PSK,
            test_psk, ARRAY_SIZE(test_psk));

    if (err) {
        LOG_ERR("Failed to load psk: %d - %s", err, nrf_err_str(err));
    }

    /* Provision Public Certificate. */
    err = nrf_inbuilt_key_write(
              sec_tag, NRF_KEY_MGMT_CRED_TYPE_IDENTITY,
              CONFIG_DEMO_MQTT_CLIENT_ID, strlen(CONFIG_DEMO_MQTT_CLIENT_ID));

    if (err) {
        LOG_ERR("Failed to load client_id: %d - %s", err, nrf_err_str(err));
    }

    return err;
}

static int tls_init(void)
{
    int err = -EINVAL;

    nrf_sec_tag_t sec_tag = (psk_enabled) ? M2MD_PSK_SEC_TAG : M2MD_CERT_SEC_TAG;

    if (tls_enabled) {
        if (psk_enabled) {
            LOG_INF("TLS-PSK Transport init");
            provision_psk_nrf(sec_tag);
        }
        else {
            LOG_INF("TLS Cert Transport init");
            provision_certificates(sec_tag);
        }
    }

    return err;
}

/**@brief Initialize the MQTT client structure
 */
static void client_init(struct mqtt_client *client)
{
    broker_init();

    mqtt_client_init(client);

    /* MQTT client configuration */
    client->broker = &broker;
    client->evt_cb = mqtt_evt_handler;
    client->client_id.utf8 = (u8_t *)CONFIG_DEMO_MQTT_CLIENT_ID;
    client->client_id.size = strlen(CONFIG_DEMO_MQTT_CLIENT_ID);
    client->password = NULL;
    client->user_name = NULL;
    client->protocol_version = MQTT_VERSION_3_1_1;

    /* MQTT buffers configuration */
    client->rx_buf = rx_buffer;
    client->rx_buf_size = sizeof(rx_buffer);
    client->tx_buf = tx_buffer;
    client->tx_buf_size = sizeof(tx_buffer);

    if (tls_enabled) {
        LOG_INF("Setting transport to tls");
        /* MQTT transport configuration */
        client->transport.type = MQTT_TRANSPORT_SECURE;

        static sec_tag_t m_sec_tags[1];

        if(psk_enabled) {
            m_sec_tags[0] = PSK_SEC_TAG;
        }
        else {
            m_sec_tags[0] = CERT_SEC_TAG;
        }

        struct mqtt_sec_config *tls_config = &client->transport.tls.config;
        tls_config->peer_verify = 0;
        tls_config->cipher_list = NULL;
        tls_config->sec_tag_list = m_sec_tags;
        tls_config->sec_tag_count = ARRAY_SIZE(m_sec_tags);
        tls_config->hostname = NULL;

    } else {
        LOG_INF("Setting transport to tcp");
        client->transport.type = MQTT_TRANSPORT_NON_SECURE;
    }
}

/**@brief Initialize the file descriptor structure used by poll.
 */
static int fds_init(struct mqtt_client *c)
{
    fds.fd = c->transport.tls.sock;
    fds.events = POLLIN;
    return 0;
}

void main(void)
{
    LOG_DBG("Mqtt Begin");
    tls_init();
    client_init(&client);

    LOG_DBG("Attempting MQTT connect");
    int err = mqtt_connect(&client);
    if (err) {
        LOG_ERR("Mqtt Connect Failed, err = %d", err);
    } else {
        LOG_DBG("Mqtt Connect Success");
    }

    err = fds_init(&client);
    if (err != 0) {
        LOG_ERR("fds_init failed: %d", err);
        return;
    }

    while (1) {
        err = poll(&fds, 1, K_SECONDS(CONFIG_MQTT_KEEPALIVE));
        if (err < 0) {
            LOG_ERR("ERROR: poll %d", errno);
            break;
        }

        err = mqtt_live(&client);
        if (err != 0) {
            LOG_ERR("ERROR: mqtt_live %d", err);
            break;
        }

        if ((fds.revents & POLLIN) == POLLIN) {
            err = mqtt_input(&client);
            if (err != 0) {
                LOG_ERR("ERROR: mqtt_input %d", err);
                break;
            }
        }

        if ((fds.revents & POLLERR) == POLLERR) {
            LOG_ERR("POLLERR");
            break;
        }

        if ((fds.revents & POLLNVAL) == POLLNVAL) {
            LOG_ERR("POLLNVAL");
            break;
        }
    }

    LOG_INF("Disconnecting MQTT client...");

    err = mqtt_disconnect(&client);
    if (err) {
        LOG_ERR("Could not disconnect MQTT client. Error: %d", err);
    }
}
